@@ -1,24 +1,28 @@
+##! 
+##! Coptright(c) 2022, 2023 Stanford Research Systems, All right reserved
+##! Subject to the MIT License
+##! 
 
 import sys
 import traceback
 import logging
 import time
 
-from matplotlib.figure import Figure
+try:
+    from matplotlib.figure import Figure
+except (ImportError, ModuleNotFoundError):
+    msg = "\n\nPython package 'Matplotlib' is required to use Task class." \
+          "\nTry again after installing 'Matplotlib' with" \
+          "\n\npip install matplotlib" \
+          "\n\nOr your system may have a different way to install it."
+    raise ModuleNotFoundError(msg)
 
-from .inputs import FloatInput, InstrumentInput
+from .inputs import FloatInput, StringInput
 from .taskresult import TaskResult, ResultLogHandler
 from .callbacks import Callbacks
 
 from srsgui.inst.instrument import Instrument
 
-# HTML formatter for QTextBrowser
-Bold = '<font color="black"><b>{}</b></font>'
-GreenBold = '<font color="green"><b>{}</b></font>'
-GreenNormal = '<font color="green">{}</font>'
-RedBold = '<font color="red"><b>{}</b></font>'
-RedNormal = '<font color="red">{}</font>'
-
 try:
     from srsgui.ui.qt.QtCore import QThread
     thread_class = QThread
@@ -26,54 +30,76 @@ except (ImportError, ModuleNotFoundError):
     from threading import Thread
     thread_class = Thread
 
+# HTML formatter for QTextBrowser
+Bold = '<font color="black"><b>{}</b></font>'
+GreenBold = '<font color="green"><b>{}</b></font>'
+GreenNormal = '<font color="green">{}</font>'
+RedBold = '<font color="red"><b>{}</b></font>'
+RedNormal = '<font color="red">{}</font>'
+
 
 class Task(thread_class):
-    """ Base class for all task classes
+    """
+    Base class for derived Task subclasses.
+
+    The parent process starts a task instance
+    as a separate thread. Before starting the task thread, the parent process injects resources
+    the task thread dependent on (instruments, figures, session handler), and connects
+    callback functions to handle the requests from the task. The task uses
+    the resources and send requests without knowing much about how they are handled by the parent
+    process. It makes easy to write a derived task as a simple Python script starting from
+    the base class.
     """
 
     class TaskException(Exception): pass
     class TaskSetupFailed(TaskException): pass
     class TaskRunFailed(TaskException): pass
+    """
+    Exceptions for Task
+    """
 
-    # When parent is not None, parent should have these attributes.
-    InstrumentDictName = 'inst_dict'  # all instruments to use in a task
-    FigureDictName = 'figure_dict'  # all Matplotlib figures to use in a task
-    SessionHandlerName = 'session_handler'
-
-    # Escape string use in stdout redirection to GUI
     EscapeForResult = '@RESULT@'
     EscapeForDevice = '@DEVICE@'
     EscapeForStatus = '@STATUS@'
     EscapeForStart = '@START@'
     EscapeForStop = '@STOP@'
+    """
+    Escape strings used in stdout redirection to GUI
+    """
 
-    # input_parameters values can be changed interactively from GUI
     input_parameters = {
         # Example float parameter
-        "Define variables before use!!": FloatInput(10.0, " Hz", 1.0, 1000.0, 1.0),
-        "Constant value": FloatInput(1.0)
+            "Define parameters": FloatInput(10.0, " Hz", 1.0, 1000.0, 1.0),
+            "before use!! ":     StringInput(" or empty input_parameters.")
     }
+    """
+    Class variable to define parameters used in the task.
+    values in input_parameters can be changed interactively from GUI before the task runs    
+    IntegerInput, FloatInput, StringInput, ListInput and InstrumentInput can be used 
+    as dictionary values.     
+    """
 
-    # Names for multiple  Matplotlib figures you will use in this task
-    # If empty, you will have one figure named 'Figure' as a default
     additional_figure_names = []  # e.g., ['Scan Plots','Temperature Plots']
+    """
+    Names for extra Matplotlib figures added to use in the task
+    If empty, only one figure named 'plot' is available as a default.
+    """
 
     _is_running = False  # class wide flag to tell if any instance is running
-    _is_optional = False  # result status will set as success initially if a task class is optional
 
-    # Image information for parent to display instead of the default logo image before instantiated
     InitialImage = None  # None for default image
+    """
+    Image file for parent to display instead of the default logo image when the task is selected.
+    """
 
     def __init__(self, parent=None):
-        """ parent should have instrument dict,
-        """
         super().__init__()
         self.parent = parent
 
-        self._keep_running = False
+        self._keep_running = False  # flag to tell if the task instance is running
         self._aborted = False
         self._error_raised = False
-        self._task_passed = False
+        self._task_passed = True
         self._log_error_detail = False  # Enables logging traceback information
 
         self.name = 'Base Task'
@@ -118,6 +144,10 @@ class Task(thread_class):
         self.logger.info('Task.cleanup() is not overridden.')
 
     def get_logger(self, name):
+        """
+        Get a logger with its handler available from the parent
+        """
+
         if self.logger_prefix:
             n = f'{self.logger_prefix}.{name}'
         else:
@@ -128,6 +158,10 @@ class Task(thread_class):
         return logger
 
     def basic_setup(self):
+        """
+        basic_setup() runs before task-specific setup()
+        """
+
         self.logger = self.get_logger(__name__)
         if self.figure is None or not hasattr(self.figure, 'canvas'):
             raise AttributeError('Invalid figure')
@@ -142,8 +176,7 @@ class Task(thread_class):
         self._keep_running = True
         self._error_raised = False
 
-        # We want to create self.result after self.name is assigned
-        self.result = TaskResult(self.name)
+        self.result = TaskResult(self.__class__.__name__)
         self.result.set_start_time_now()
 
         log_format = '%(asctime)s-%(levelname)s-%(message)s'
@@ -160,9 +193,13 @@ class Task(thread_class):
         self.logger.info(GreenBold.format(msg))
 
         self._notify_start()
-        self.clear()
+        self.clear_figures()
 
     def basic_cleanup(self):
+        """
+        basic_cleanup runs after task-specific cleanup()
+        """
+
         try:
             Task._is_running = False
             self._keep_running = False
@@ -176,7 +213,7 @@ class Task(thread_class):
                 self.logger.info(msg)
                 self.result.set_aborted()
             elif self._task_passed:
-                msg = '{} PASSED'.format(self.name)
+                msg = '{} FINISHED'.format(self.name)
                 self.display_result(msg)
                 self.update_status(msg)
                 self.logger.info(GreenBold.format(msg))
@@ -187,13 +224,19 @@ class Task(thread_class):
                 self.update_status(msg)
                 self.logger.info(RedBold.format(msg))
                 self.result.set_passed(False)
+
+            if self.session_handler:
+                self.session_handler.create_new_task_result(self.result)
+                self.logger.debug('A task result for DUT is created')
         except Exception as e:
             self.logger.error('Error during basic_cleanup: {}'.format(e))
         finally:
-            # self.logger.removeHandler(self.file_log_handler)
             self.logger.removeHandler(self.result_log_handler)
 
     def run(self):
+        """
+        Overrides Thread run() method. task-speciic test() runs inside this method.
+        """
         try:
             self.basic_setup()
             if self._keep_running:
@@ -217,25 +260,44 @@ class Task(thread_class):
         except Exception as e:
             self.log_exception(e)
 
-    # Override for QThread start
     def start(self):
-        # if Task._is_running: # Disable for multiple tasks running
-        #     raise RuntimeError('Another task is running')
+        """
+        Overrides Thread start() method.
+        """
 
         self._keep_running = True
         self._aborted = False
         super().start()
-        # self.logger.debug("{} start completed".format(self.name))
 
     def stop(self):
+        """
+        Make is_running() returns False. A task should check is_running()
+        frequently. Stop if it returns False.
+        """
+
         if self._keep_running:
             self._aborted = True
             self._keep_running = False
 
+    def delay(self, seconds):
+        """
+        Check if the task is stopped and wait for the given seconds.
+        """
+        if not self._keep_running:
+            raise Task.TaskException('Task is requested to stop')
+        else:
+            time.sleep(seconds)
+
     def set_session_handler(self, session_handler):
+        """
+        Parent should set a session handler for Task to use file output.
+        """
         self.session_handler = session_handler
 
     def set_callback_handler(self, callback_handler: Callbacks):
+        """
+        Parent should set a callback handler to handle events from Task.
+        """
         self.callbacks = callback_handler
 
     @staticmethod
@@ -248,11 +310,26 @@ class Task(thread_class):
         return True
 
     def set_inst_dict(self, inst_dict):
+        """
+        Parent should set inst_dict for Task to use instruments available from the parent.
+        """
+
         if not self._check_dict_items(inst_dict, Instrument):
             raise AttributeError('invalid inst_dict for Task class')
         self.inst_dict = inst_dict
 
+    def set_data_dict(self, data_dict):
+        """
+        A dictionary injected when the task run. It is a way to share data among different tasks
+        Do not reset the whole dictionary unless you know what you are doing.
+        """
+        self.data_dict = data_dict
+
     def set_figure_dict(self, figure_dict):
+        """
+        Parent should set figure_dict for Task to use Matplotlib figures available from the parent.
+        """
+
         if not self._check_dict_items(figure_dict, Figure):
             raise AttributeError('invalid figure_dict for Task class')
         self.figure_dict = figure_dict
@@ -263,28 +340,33 @@ class Task(thread_class):
             raise ValueError('No figure in figure_dict to set as default')
 
     def get_figure(self, name=None) -> Figure:
+        """
+        Get a Matplotlib figure from figure_dict.
+        if name is None, it will reutrn the first figure in figure_dict as the defualt
+        """
+
         if name is None:
             name = list(self.figure_dict.keys())[0]
         if name in self.figure_dict:
             return self.figure_dict[name]
         raise KeyError('Invalid figure name: {}'.format(name))
 
-    def clear(self):
+    def clear_figures(self):
         """
-        Clear figures
+        Clear all the figures in figure_dict
         """
         for fig in self.figure_dict.values():
             if hasattr(fig, 'canvas'):
                 fig.clear()
-                fig.canvas.draw_idle()
+                self.request_figure_update(fig)
 
     def is_running(self):
+        """
+        Task should check is_running() is True.
+        If it returns False, Task should stop ASAP.
+        """
         return self._keep_running
 
-    @classmethod
-    def is_optional(cls):
-        return cls._is_optional
-
     def is_task_passed(self):
         return self._task_passed
 
@@ -295,10 +377,15 @@ class Task(thread_class):
         self.result.set_passed(status)
 
     def is_error_raised(self):
+        """
+        Check if Task stopped with an error
+        """
         return self._error_raised
 
-    # Set to stop the task selection from running further
     def set_error_raised(self):
+        """
+        Mark Task is stopped with an error
+        """
         self._error_raised = True
 
     # Wrapper for TaskResult.add_details
@@ -328,6 +415,14 @@ class Task(thread_class):
             raise AttributeError("No session handler available")
         self.session_handler.add_to_table_in_file(*args, format_list=format_list)
 
+    def add_dict_to_file(self, name, data_dict):
+        """
+        Add a dictionary to the file.
+        """
+        if self.session_handler is None:
+            raise AttributeError("No session handler available")
+        self.session_handler.add_dict_to_file(name, data_dict)
+
     def save_result(self, msg):
         self.logger.error('Do not use save_result. Use add_to_table_in_file, or result.add_details, instead.')
         # raise NotImplementedError()
@@ -339,6 +434,7 @@ class Task(thread_class):
 
     # Following methods will be used with stdout redirection
     def update_status(self, message):
+        """ Output to the status bar at the bottom of srsgui application"""
         self.write_text(self.EscapeForStatus + message)
 
     def display_device_info(self, message='', device_name=None, update=False,  clear=False):
@@ -385,10 +481,7 @@ class Task(thread_class):
     def get_input_parameter(self, name):
         if name in self.__class__.input_parameters:
             param = self.__class__.input_parameters[name]
-            if type(param) == InstrumentInput:
-                value = self.__class__.input_parameters[name].text
-            else:
-                value = self.__class__.input_parameters[name].value
+            value = param.get_value()
             if not hasattr(self.result, name):
                 self.add_details(str(value), name)
             return value
@@ -405,18 +498,7 @@ class Task(thread_class):
     @classmethod
     def set_input_parameter(cls, name, value):
         if name in cls.input_parameters:
-            param = cls.input_parameters[name]
-            if type(param) == InstrumentInput:
-                if type(cls.input_parameters[name].text) == type(value):
-                    cls.input_parameters[name].text = value
-                else:
-                    raise TypeError('Type for input_parameter "{}" is {}'
-                                    .format(name, type(value)))
-            elif type(cls.input_parameters[name].value) == type(value):
-                cls.input_parameters[name].value = value
-            else:
-                raise TypeError('Type for input_parameter "{}" is {}'
-                                .format(name, type(value)))
+            cls.input_parameters[name].set_value(value)
         else:
             raise KeyError('{} not in input_parameters'.format(name))
 
@@ -432,26 +514,20 @@ class Task(thread_class):
     def update_figure(self, figure: Figure):
         if type(figure) is not Figure:
             raise TypeError('{} is not  a Figure'.format(type(figure)))
-        figure.canvas.draw_idle()
+        self.callbacks.figure_update_requested(figure)
 
     # It needs a matching update() as a slot to run from UI
-    def notify_data_available(self, data_dict={}):
-        self.callbacks.data_available(data_dict)
-
-    # These callbacks are used to update display for streaming data from another class or thread
-    # Signals are wrapped as a callback functions 
-
-    def data_available_callback(self, data_dict={}, *args):
-        self.callbacks.data_available(data_dict)
+    def notify_data_available(self, data={}):
+        self.callbacks.data_available(data)
 
     def update(self, data: dict):
         """
-        when data_available signal emits, this method handles new data.
-        By default, it only updates the matplotlib figure.
+        when notify_data_available is called, this method handles data processing
+        and display update. By default, it does no data handling, but figure update request.
+        GUI related data processing needs to be done here to be handled
+        in proper order by the GUI event loop handler.
         """
         self.request_figure_update()
-        # self.logger.error("Derive update() to use data_available signal")
-        # raise NotImplementedError("Derive update() to use data_available signal")
 
     def get_instrument(self, name):
         """Get an instrument from parent's inst_dict and check its validity"""
@@ -506,14 +582,19 @@ class Task(thread_class):
         raise Task.TaskRunFailed("Timeout at '{}'".format(self.current_question))
 
     def question_background_update(self):
+        """Repeat this method, while a dialog box is open with ask_question method.
+        A Task subclass can re-implement this method as it needs"""
         time.sleep(0.1)
 
     def set_log_error_detail(self, state=False):
+        """Set True To log exception traceback for debugging
+        """
         self._log_error_detail = state
 
     def log_exception(self, err):
-        # Capture the error
-        self.logger.error(err)
+        """With set_log_error_detail(True), an error is logged with traceback information
+        """
+        self.logger.error(f'{err.__class__.__name__}: {err}')
 
         if self._log_error_detail:
             # And the stack trace
