@@ -1,3 +1,7 @@
+##! 
+##! Coptright(c) 2022, 2023 Stanford Research Systems, All right reserved
+##! Subject to the MIT License
+##! 
 
 import sys
 
@@ -5,8 +9,13 @@ from .qt.QtCore import QObject
 from .qt.QtCore import Signal
 
 
-# redirect STDOUT and STDERR to a function
 class StdOut(QObject):
+    """
+    Redirect the system standard output and standard error output to
+    the output_function in __init__().
+    It is instantiated with :meth:`srsgui.ui.taskmain.TaskMain.print_redirect`
+    for the main application.
+    """
     textWritten = Signal(str)
 
     def __init__(self, output_function, parent=None):
