@@ -1,3 +1,7 @@
+##! 
+##! Coptright(c) 2022, 2023 Stanford Research Systems, All right reserved
+##! Subject to the MIT License
+##! 
 
 import os
 import sys
@@ -7,32 +11,43 @@ import logging.handlers
 
 from pathlib import Path
 
+from .qt import QT_BINDER, PYSIDE6, QT_BINDER_VERSION
 from .qt.QtCore import QTimer, QSettings
 from .qt.QtWidgets import QMainWindow, QApplication, QTextBrowser,\
                                 QVBoxLayout, QMessageBox, \
                                 QInputDialog, QFileDialog, \
-                                QAction
+                                QMenu, QAction
 
 from .ui_taskmain import Ui_TaskMain
-from .commConnectDlg import CommConnectDlg
+
+from .connectdlg import ConnectDlg
+
 from .inputpanel import InputPanel
 from .signalhandler import SignalHandler
 
 from .stdout import StdOut
 from .qtloghandler import QtLogHandler
 from .deviceinfohandler import DeviceInfoHandler
+
 from .dockhandler import DockHandler
 
 from srsgui.task.config import Config
 from srsgui.task.sessionhandler import SessionHandler
-from srsgui.task import Task, Bold
-from srsgui.inst.instrument import Instrument
+from srsgui.task.task import Task, Bold
+
+from srsgui import __version__
 
 logger = logging.getLogger(__name__)
 
 
 class TaskMain(QMainWindow, Ui_TaskMain):
-    DefaultConfigFile = "rga.taskconfig"
+    """
+    The main window of the SRSGUI application
+    """
+
+    DefaultConfigFile = str(Path(__file__).parent.parent /
+        "examples/oscilloscope example/oscilloscope example project.taskconfig")
+
     OrganizationName = 'srsinst'
     ApplicationName = 'srsgui'
 
@@ -50,6 +65,7 @@ class TaskMain(QMainWindow, Ui_TaskMain):
 
         # The dict holds subclass of Task
         self.task_dict = {}
+        self.task_menus = []
         self.current_task_action = None
         self.task = None
         self.task_method = None
@@ -57,17 +73,21 @@ class TaskMain(QMainWindow, Ui_TaskMain):
         self.question_result = None
         self.question_result_value = None
 
+        # data_dict hold data shared among task. It will inject to a task when run
+        self.data_dict = {}
+
         # self.inst_dict holds instances of subclass of Instrument
         self.inst_dict = {}
         self.inst_info_handler = DeviceInfoHandler(self)
 
         self.dock_handler = DockHandler(self)
+        self.command_handler = self.dock_handler.terminal_command_handler
         self.console = self.dock_handler.console
         self.terminal_widget = self.dock_handler.terminal_widget
-        self.figure = self.dock_handler.get_figure()
-        self.figure_dict = self.dock_handler.get_figure_dict()
         self.plotDockWidget = self.dock_handler.get_dock()
 
+        # Make the terminal not blocking for log query
+
         self.geometry_dict = {}
         try:
             default_config_file = self.DefaultConfigFile
@@ -119,6 +139,14 @@ class TaskMain(QMainWindow, Ui_TaskMain):
         self.statusbar.showMessage('Waiting for task selection')
         self.stdout = StdOut(self.print_redirect)
 
+        self.about = QAction(self)
+        self.about.setText('About')
+        self.menu_Help.addAction(self.about)
+        self.about.triggered.connect(self.onAbout)
+
+        self.menu_Tasks.triggered.connect(self.onTaskSelect)
+        self.menu_Instruments.triggered.connect(self.onInstrumentSelect)
+
     def load_tasks(self):
         try:
             # Clear console and result display
@@ -158,9 +186,10 @@ class TaskMain(QMainWindow, Ui_TaskMain):
                 del instr
             self.inst_dict = self.config.inst_dict
 
+            self.dock_handler.reset_inst_docks()
+
             self.inst_info_handler.update_tabs()
             for inst_name in self.inst_dict:
-                inst = self.get_inst(inst_name)
                 self.inst_info_handler.update_info(inst_name)
 
             self.task_dict = self.config.task_dict
@@ -176,58 +205,64 @@ class TaskMain(QMainWindow, Ui_TaskMain):
             logger.error('{}: {}'.format(e.__class__.__name__, e))
 
         try:
-            try:
-                # Disconnect with none connected causes an exception
-                self.menu_Instruments.triggered.disconnect()
-            except:
-                pass
             actions = self.menu_Instruments.actions()
             for action in actions:
                 self.menu_Instruments.removeAction(action)
-
             for item in self.inst_dict:
                 action_inst = QAction(self)
                 action_inst.setText(item)
                 self.menu_Instruments.addAction(action_inst)
-
-                # action_inst.setCheckable(True)
-                # if hasattr(self.inst_dict[item], 'is_connected') and self.inst_dict[item].is_connected():
-                #    action_inst.setChecked(True)
-            self.menu_Instruments.triggered.connect(self.onInstrumentSelect)
-
-            try:
-                self.menu_Tasks.triggered.disconnect()
-            except:
-                pass
-
+            logger.debug('Added new actions to Instruments menu')
+            # Remove previous actions from Task menu
             actions = self.menu_Tasks.actions()
             for action in actions:
                 self.menu_Tasks.removeAction(action)
 
-            for item in self.task_dict:
+            # Add new actions to Task menu
+            for name in self.task_dict:
+                m = self.menu_Tasks
+
+                # set up submenu structure
+                p = self.config.task_path_dict[name]
+                if p:
+                    tokens = p.split('/')
+                    for token in tokens:
+                        exists = False
+                        na = None
+                        if type(m) == QAction:
+                            ma = m.menu()
+                            if ma:
+                                m = ma
+                            else:
+                                continue
+                        for action in m.actions():
+                            if token == action.text():
+                                na = action
+                                exists = True
+                                break
+                        if not exists:
+                            na = m.addMenu(QMenu(token, m))
+                        m = na
+                    if type(m) == QAction:
+                        ma = m.menu()
+                        if ma:
+                            m = ma
+                        else:
+                            continue
                 action_task = QAction(self)
-                action_task.setText(item)
-                self.menu_Tasks.addAction(action_task)
-            self.menu_Tasks.triggered.connect(self.onTaskSelect)
+                action_task.setText(name)
+                m.addAction(action_task)
         except Exception as e:
-            print(e)
+            logger.error('Error adding to Task menu Task:{}  Error: {}'.format(name, e))
 
     def get_logo_file(self):
         return self.LogoFile
 
-    def onInstrumentSelect(self, inst_action):
-        try:
-            name = inst_action.text()
-            self.inst_info_handler.select_browser(name)
-            if self.inst_dict[name].is_connected():
-                self.inst_info_handler.update_info(name)
-                self.onDisconnect(name)
-            else:
-                self.onConnect(name)
-        except Exception as e:
-            logger.error(e)
-
     def print_redirect(self, text):
+        """
+        Handles text output for stdout, stderr and various text output
+        from :class:`srsgui.task.task.Task`
+        """
         try:
             if len(text) < 2:
                 return
@@ -286,10 +321,9 @@ class TaskMain(QMainWindow, Ui_TaskMain):
         self.actionStop.setEnabled(True)
 
         self.taskResult.clear()
-        self.dock_handler.show_toolbar(True)
 
         self._busy_flag = True
-
+        self.dock_handler.show_toolbar(True)
         self.session_handler.create_file(self.task.__class__.__name__)
 
     def onTaskFinished(self):
@@ -298,15 +332,13 @@ class TaskMain(QMainWindow, Ui_TaskMain):
             # Setup toolbar buttons
             self.actionRun.setEnabled(True)
             self.actionStop.setEnabled(False)
-
+            self.setWindowTitle(self.config.task_dict_name)
             self._busy_flag = False
 
             # PySide2 needs to refresh matplotlib display before starting a new task
             # for live update. Hide and show toolbar does the trick.
             self.dock_handler.show_toolbar(False)
-
-            self.create_task_result_in_session(self.task)
-
+            self.session_handler.close_file()
             # try:
             #     self.task.deleteLater()
             # except Exception as e:
@@ -315,18 +347,15 @@ class TaskMain(QMainWindow, Ui_TaskMain):
 
             for inst in self.inst_dict:
                 self.inst_info_handler.update_info(inst)
-
+            logger.debug('onTaskFinished finished')
         except Exception as e:
             logger.error('Error onTaskFinished: {}'.format(e))
 
     def is_task_running(self):
         return self._busy_flag
 
-    def get_inst(self, inst):
-        if inst in self.inst_dict:
-            return self.inst_dict[inst]
-        else:
-            return None
+    def update_figure(self, figure):
+        self.dock_handler.update_figure(figure)
 
     def onTaskSelect(self, action):
         try:
@@ -353,17 +382,22 @@ class TaskMain(QMainWindow, Ui_TaskMain):
                 raise TypeError(msg)
 
             self.task_method = taskClassChosen
-
             self.statusbar.showMessage('Press Run button to start the task selected')
-
-            self.taskResult.setText(self.task_method.__doc__)
+            if QT_BINDER == PYSIDE6:  # To bypass PySide6 docstring bug
+                task = self.task_method()
+                doc = task.__doc__
+                del task
+            else:
+                doc = self.task_method.__doc__
+            self.taskResult.setText(doc)
             self.taskParameterFrame.layout().removeWidget(self.taskParameter)
             self.taskParameter.deleteLater()
             self.taskParameter = InputPanel(self.task_method, self)
             self.taskParameterFrame.layout().addWidget(self.taskParameter)
+            self.label_task_params.setText('Parameters in  {}'.format(current_action_name))
+            self.label_task_params.setWordWrap(True)
 
-            self.dock_handler.update_figures(self.task_method.additional_figure_names)
-            self.figure_dict = self.dock_handler.get_figure_dict()
+            self.dock_handler.reset_figures(self.task_method.additional_figure_names)
 
             self.dock_handler.show_toolbar(False)
             self.handle_initial_image(self.task_method)
@@ -390,13 +424,13 @@ class TaskMain(QMainWindow, Ui_TaskMain):
 
             self.task = self.task_method(self)
             self.task.name = self.current_task_action.text()
+            self.setWindowTitle("{}  -  {}".format(self.config.task_dict_name, self.task.name))
             self.task.set_figure_dict(self.dock_handler.get_figure_dict())
             self.task.set_inst_dict(self.inst_dict)
+            self.task.set_data_dict(self.data_dict)
             self.task.set_session_handler(self.session_handler)
-
             signal_handler = SignalHandler(self)
             self.task.set_callback_handler(signal_handler)
-
             self.onTaskStarted()
             self.task.start()
         except Exception as e:
@@ -419,11 +453,30 @@ class TaskMain(QMainWindow, Ui_TaskMain):
         except Exception as e:
             logger.error(e)
 
+    def onInstrumentSelect(self, inst_action):
+        try:
+            name = inst_action.text()
+            self.inst_info_handler.select_browser(name)
+            if self.inst_dict[name].is_connected():
+                try:
+                    self.inst_info_handler.update_info(name)
+                except:
+                    pass
+                self.onDisconnect(name)
+            else:
+                self.onConnect(name)
+        except Exception as e:
+            logger.error(e)
+
     def onConnect(self, inst_name):
+        if inst_name not in self.inst_dict:
+            raise KeyError('Invalid instrument name: {}'.format(inst_name))
+
         logger.info('Connecting to {}...'.format(inst_name))
         try:
-            inst = self.get_inst(inst_name)
-            form = CommConnectDlg(inst)
+            inst = self.inst_dict[inst_name]
+            form = ConnectDlg(inst)
+            # form = CommConnectDlg(inst)
             form.exec_()
 
             if inst.is_connected():
@@ -435,13 +488,21 @@ class TaskMain(QMainWindow, Ui_TaskMain):
             logger.error(e)
 
     def onDisconnect(self, inst_name):
-        inst = self.get_inst(inst_name)
-        if inst.is_connected():
+        try:
+            if not (inst_name in self.inst_dict and hasattr(self.inst_dict[inst_name], "is_connected")):
+                raise KeyError('Invalid instrument: {}'.format(inst_name))
+
+            inst = self.inst_dict[inst_name]
+            if not self.inst_dict[inst_name].is_connected():
+                logger.error('{} is not connected'.format(inst_name))
+                return
             self.display_question("Do you want to disconnect '{}'?".format(inst_name))
             if self.question_result_value:
                 inst.disconnect()
                 self.inst_info_handler.update_info(inst_name)
                 logger.info('{} is disconnected'.format(inst_name))
+        except Exception as e:
+            logger.error(e)
 
     def okToContinue(self):
         if self.task and self.task.is_running():
@@ -450,14 +511,6 @@ class TaskMain(QMainWindow, Ui_TaskMain):
                 return False
         return True
 
-    def create_task_result_in_session(self, task):
-        if not self.session_handler.is_open():
-            logger.error('No session is open when the task is finished')
-            return
-        self.session_handler.create_new_task_result(task.result)
-        self.session_handler.close_file()
-        logger.debug('A task result for DUT is created')
-
     def display_question(self, question, return_type=bool):
         try:
             self.question_result = None
@@ -508,10 +561,12 @@ class TaskMain(QMainWindow, Ui_TaskMain):
             # Save settings
             self.save_settings()
 
+            self.dock_handler.stop_command_handlers()
+
             # Close instruments
             inst_dict = self.inst_dict
             for key in inst_dict:
-                if issubclass(type(inst_dict[key]), Instrument):
+                if hasattr(inst_dict[key], "disconnect"):
                     inst_dict[key].disconnect()
         else:
             event.ignore()
@@ -530,6 +585,21 @@ class TaskMain(QMainWindow, Ui_TaskMain):
         except Exception as e:
             logger.error(f"Error in handle_initial_image: {e}")
 
+    def onAbout(self,checked):
+        msg = ''
+        for name in self.inst_dict:
+            inst = self.inst_dict[name]
+            if hasattr(inst, __version__):
+                version = inst.__version__
+            else:
+                version = 'N/A'
+
+            msg += '{} version: {}\n'.format(inst.__class__.__name__, version)
+        msg += '\nSrsgui version: {}\n'.format(__version__)
+        msg += '\n{} version: {}\n'.format(QT_BINDER, QT_BINDER_VERSION)
+        msg += 'Python version: {}\n'.format(sys.version)
+        self.display_question(msg, None)
+
     def load_settings(self):
         try:
             self.default_config_file = self.settings.value("ConfigFile", "")
@@ -537,8 +607,11 @@ class TaskMain(QMainWindow, Ui_TaskMain):
             self.splitter.setSizes([int(i) for i in sizes])
             sizes = self.settings.value("MainWindow/Splitter2", [150, 500])
             self.splitter_2.setSizes([int(i) for i in sizes])
-            self.restoreGeometry(self.settings.value("MainWindow/Geometry"))
-            self.restoreState(self.settings.value("MainWindow/State"))
+            geo = self.settings.value("MainWindow/Geometry")
+            state = self.settings.value("MainWindow/State")
+            if geo and state:
+                self.restoreGeometry(geo)
+                self.restoreState(state)
         except Exception as e:
             logger.error('During load_setting, {}'.format(e))
 
